#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os
import time
import uuid
import yaml

import six
from six.moves import http_client
from six.moves.configparser import SafeConfigParser

from gssapi.exceptions import GSSError
from ipalib import api
from ipalib import errors

try:
    from ipapython.ipautil import kinit_keytab
except ImportError:
    # The import moved in freeIPA 4.5.0
    from ipalib.install.kinit import kinit_keytab

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.openstack.cloud.plugins.module_utils.openstack import openstack_full_argument_spec
from ansible_collections.openstack.cloud.plugins.module_utils.openstack import openstack_module_kwargs

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'status': ['preview'],
    'supported_by': 'community'
}

DOCUMENTATION = '''
---
module: add_dns_entries

short_description: Add DNS entries and zones

version_added: "2.8"

description:
    - Adds DNS forward and reverse entries in a batch operation.

options:
    keytab:
        description:
            - Keytab to use when authenticating to FreeIPA
        type: str
    zones:
        description:
            - DNS zones to be added
        type: list
    records:
        description:
            - DNS forward records to be added
        type: list
author:
    - Ade Lee (@vakwetu)
'''

EXAMPLES = '''
- name: Add DNS records and zones
  add_dns_entries:
      keytab: /etc/krb5.keytab
      zones:
        - ctlplane.example.com
        - storage.example.com
        - 1.168.192.in-addr.arpa
        - 0.0.0.0.1.0.0.0.8.b.d.0.1.0.0.2.ip6.arpa
      records:
      - []
'''


class IPAClient(object):

    def __init__(self, keytab):
        self.ntries = 5
        self.retry_delay = 2
        self.keytab = keytab

        if self._ipa_client_configured() and not api.isdone('finalize'):
            self.ccache = "MEMORY:" + str(uuid.uuid4())
            os.environ['KRB5CCNAME'] = self.ccache
            (hostname, realm) = self.get_host_and_realm()
            self.realm = realm
            kinit_keytab(str('nova/%s@%s' % (hostname, realm)),
                         self.keytab, self.ccache)
            api.bootstrap(context='cleanup')
            api.finalize()
        else:
            self.ccache = os.environ['KRB5CCNAME']
        self.batch_args = list()

    def get_host_and_realm(self):
        """Return the hostname and IPA realm name."""
        config = SafeConfigParser()
        config.read('/etc/ipa/default.conf')
        hostname = config.get('global', 'host')
        realm = config.get('global', 'realm')
        return (hostname, realm)

    def __get_connection(self):
        """Make a connection to IPA or raise an error."""
        tries = 0

        while (tries <= self.ntries):
            logging.debug("Attempt %d of %d", tries, self.ntries)
            if api.Backend.rpcclient.isconnected():
                api.Backend.rpcclient.disconnect()
            try:
                api.Backend.rpcclient.connect()
                # ping to force an actual connection in case there is only one
                # IPA master
                api.Command[u'ping']()
            except (errors.CCacheError,
                    errors.TicketExpired,
                    errors.KerberosError) as e:
                tries += 1

                # pylint: disable=no-member
                logging.debug("kinit new ccache in get_connection: %s", e)
                try:
                    kinit_keytab(str('nova/%s@%s' %
                                 (api.env.host, api.env.realm)),
                                 self.keytab,
                                 self.ccache)
                except GSSError as e:
                    logging.debug("kinit failed: %s", e)
            except errors.NetworkError:
                tries += 1
            except http_client.ResponseNotReady:
                # NOTE(xek): This means that the server closed the socket,
                # so keep-alive ended and we can't use that connection.
                api.Backend.rpcclient.disconnect()
                tries += 1
            else:
                # successful connection
                return
            logging.debug("Waiting %s seconds before next retry.",
                          self.retry_delay)
            time.sleep(self.retry_delay)

        logging.error(" Failed to connect to IPA after %d attempts",
                      self.ntries)
        raise Exception("Failed to connect to IPA")

    def start_batch_operation(self):
        """Start a batch operation.

           IPA method calls will be collected in a batch job
           and submitted to IPA once all the operations have collected
           by a call to _flush_batch_operation().
        """
        logging.debug("start batch operation")
        self.batch_args = list()

    def _add_batch_operation(self, command, *args, **kw):
        """Add an IPA call to the batch operation"""
        self.batch_args.append({
            "method": six.text_type(command),
            "params": [args, kw],
        })

    def flush_batch_operation(self):
        """Make an IPA batch call."""
        logging.debug("flush_batch_operation")
        if not self.batch_args:
            return None

        kw = {}
        logging.debug(" %s", self.batch_args)

        return self._call_ipa('batch', *self.batch_args, **kw)

    def _call_ipa(self, command, *args, **kw):
        """Make an IPA call."""
        if not api.Backend.rpcclient.isconnected():
            self.__get_connection()
        if 'version' not in kw:
            kw['version'] = u'2.146'  # IPA v4.2.0 for compatibility

        while True:
            try:
                result = api.Command[command](*args, **kw)
                logging.debug(result)
                return result
            except (errors.CCacheError,
                    errors.TicketExpired,
                    errors.KerberosError):
                logging.debug("Refresh authentication")
                self.__get_connection()
            except errors.NetworkError:
                raise
            except http_client.ResponseNotReady:
                # NOTE(xek): This means that the server closed the socket,
                # so keep-alive ended and we can't use that connection.
                api.Backend.rpcclient.disconnect()
                raise

    def _ipa_client_configured(self):
        """Determine if the machine is an enrolled IPA client.

           Return boolean indicating whether this machine is enrolled
           in IPA. This is a rather weak detection method but better
           than nothing.
        """

        return os.path.exists('/etc/ipa/default.conf')

    def add_dns_zone(self, zone):
        logging.debug('Adding DNS zone: %s', zone)
        params = [zone]
        self._add_batch_operation('dnszone_add', *params)

    def add_dns_record(self, record):
        args = dict()
        params = [record['zone_name'], record['record_name']]
        record_type = record['record_type']
        rec = "{0}record".format(record_type.lower())
        args[rec] = [record['record_value']]
        self._add_batch_operation('dnsrecord_add', *params, **args)


def add_dns_zones(ipa, zones):
    ipa.start_batch_operation()
    for zone in zones:
        ipa.add_dns_zone(zone)
    ipa.flush_batch_operation()


def add_dns_records(ipa, records):
    ipa.start_batch_operation()
    for record in records:
        ipa.add_dns_record(record)
    ipa.flush_batch_operation()


def run_module():
    argument_spec = openstack_full_argument_spec(
        **yaml.safe_load(DOCUMENTATION)['options']
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
        **openstack_module_kwargs()
    )

    try:
        keytab = module.params.get('keytab')
        zones = module.params.get('zones')
        records = module.params.get('records')

        ipa = IPAClient(keytab)
        add_dns_zones(ipa, zones)
        add_dns_records(ipa, records)

        module.exit_json(changed=True)
    except Exception as err:
        module.fail_json(msg=str(err))


def main():
    run_module()


if __name__ == '__main__':
    main()
